from itertools import chain
import re

from . import APIClient, Project
from requests.exceptions import HTTPError

ANONYMOUS_USER_ID = 2

class RedmineClient(APIClient):
    PAGE_MAX_SIZE = 100

    def get_auth_headers(self):
        return {"X-Redmine-API-Key": self.api_key}

    def get(self, *args, **kwargs):
        # In detail views, redmine encapsulate "foo" typed objects under a
        # "foo" key on the JSON.
        ret = super().get(*args, **kwargs)
        values = ret.values()
        if len(values) == 1:
            return list(values)[0]
        else:
            return ret

    def unpaginated_get(self, *args, **kwargs):
        """ Iterates over API pagination for a given resource list
        """
        kwargs['params'] = kwargs.get('params', {})
        kwargs['params']['limit'] = self.PAGE_MAX_SIZE

        resp = self.get(*args, **kwargs)

        # Try to autofind the top-level key containing
        keys_candidates = (
            set(resp.keys()) - set(['total_count', 'offset', 'limit']))

        assert len(keys_candidates) == 1
        res_list_key = list(keys_candidates)[0]

        result_pages = [resp[res_list_key]]
        if 'offset' not in resp:
            raise ValueError('HTTP response data is not paginated')

        while (resp['total_count'] - resp['offset'] - resp['limit']) > 0:
            kwargs['params']['offset'] = (kwargs['params'].get('offset', 0)
                                          + self.PAGE_MAX_SIZE)
            resp = self.get(*args, **kwargs)
            result_pages.append(resp[res_list_key])
        return chain.from_iterable(result_pages)


class RedmineProject(Project):
    REGEX_PROJECT_URL = re.compile(
        r'^(?P<base_url>https?://.*)/projects/(?P<project_name>[\w_-]+)$')

    REGEX_CATEGORY_PROJECT_URL = re.compile(
        r'^(?P<base_url>https?://.*)/project/(?P<category_name>[\w_-]+)/(?P<project_name>[\w_-]+)/?$')

    def __init__(self, url, *args, **kwargs):
        normalized_url = self._canonicalize_url(url)
        super().__init__(normalized_url, *args, **kwargs)
        self.api_url = '{}.json'.format(self.public_url)
        self.instance_url = self._url_match.group('base_url')

    @classmethod
    def _canonicalize_url(cls, url):
        """ If using caterogies, return the category-less URL

        eg:
          - category URL: https://example.com/project/dev/foobar/
          - category-less URL: https://example.com/projects/foobar/

        API endpoints are reachable only for category-less URLs.
        """
        m = cls.REGEX_CATEGORY_PROJECT_URL.match(url)
        if m:
            return '{base_url}/projects/{project_name}'.format(**m.groupdict())
        else:
            return url

    def get_all_issues(self):

        if not hasattr(self, '_cache_issues'):

            issues = self.api.unpaginated_get(
                '{}/issues.json?subproject_id=1&status_id=*'.format(self.public_url))
            detailed_issues = []
            # It's impossible to get issue history from list view, so get it from
            # detail view...

            for issue_id in sorted(i['id'] for i in issues):
                issue_url = '{}/issues/{}.json?include=journals,watchers,relations,children,attachments,changesets'.format(
                    self.instance_url, issue_id)
                detailed_issues.append(self.api.get(issue_url))

            self._cache_issues = detailed_issues

        return self._cache_issues

    def get_all_pages(self):
        return self.api.get(
            '{}/wiki/index.json'.format(self.public_url))

    def get_page(self, title, version):
        return self.api.get(
            '{}/wiki/{}/{}.json'.format(self.public_url, title, version))

    def get_participants(self):
        """Get participating users (issues authors/owners)

        :return: list of all users participating on issues
        :rtype: list
        """
        user_ids = set()
        users = []

        for i in self.get_all_issues():
            journals = i.get('journals', [])
            for i in chain(i.get('watchers', []),
                           [i['author'], i.get('assigned_to', None)]):

                if i is None or ('name' not in i and 'author' not in i):
                    continue
                user_ids.add(i['id'])
            for entry in journals:
                if not entry.get('notes', None):
                    continue
                user_ids.add(entry['user']['id'])

        for i in user_ids:
            # The anonymous user is not really part of the project...
            # You may want to add Group IDs such as [ANONYMOUS_USER_ID, 324, 234, ...] if necessary
            if i not in [ANONYMOUS_USER_ID]:
                try:
                  users.append(self.api.get('{}/users/{}.json'.format(
                      self.instance_url, i)))
                except HTTPError:
                  print("unable to retrieve user!")

        return users

    def get_users_index(self):
        """ Returns dict index of users (by user id)
        """
        return {i['id']: i for i in self.get_participants()}

    def get_versions(self):
        response = self.api.get('{}/versions.json'.format(self.public_url))
        return response['versions']



import re
import logging
import requests
from . import APIClient, Project
import urllib
from urllib.request import urlopen

from redmine_gitlab_migrator.converters import redmine_username_to_gitlab_username

from json.decoder import JSONDecodeError

log = logging.getLogger(__name__)

class GitlabClient(APIClient):
    # see http://doc.gitlab.com/ce/api/#pagination
    MAX_PER_PAGE = 100

    def get(self, *args, **kwargs):
        kwargs['params'] = kwargs.get('params', {})
        kwargs['params']['page'] = 1
        kwargs['params']['per_page'] = self.MAX_PER_PAGE

        result = super().get(*args, **kwargs)
        while (len(result) > 0 and len(result) % self.MAX_PER_PAGE == 0):
            kwargs['params']['page'] += 1
            result.extend(super().get(*args, **kwargs))
        return result

    def get_auth_headers(self):
        return {"PRIVATE-TOKEN": self.api_key}

    def check_is_admin(self):
        pass


class GitlabInstance:
    def __init__(self, url, client):
        self.url = url.strip('/')  # normalize URL
        self.api = client

    def get_all_users(self):
        return self.api.get('{}/users'.format(self.url))

    def get_users_index(self):
        """ Returns dict index of users (by login)
        """
        return {i['username']: i for i in self.get_all_users()}

    def get_group_members(self, group_id):
        return self.api.get('{}/groups/{}/members'.format(self.url, group_id))


    def check_users_exist(self, usernames):
        """ Returns True if all users exist
        """
        gitlab_user_names = set([i['username'] for i in self.get_all_users()])

        translated = []
        for i in usernames:
            print(i, redmine_username_to_gitlab_username(i))
            translated.append(redmine_username_to_gitlab_username(i))
        return all((i in gitlab_user_names for i in translated))


class GitlabProject(Project):
    REGEX_PROJECT_URL = re.compile(
        r'^(?P<base_url>https?://[^/]+/)(?P<namespace>[\.\w\._/-]+)/(?P<project_name>[\w\._-]+)$')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.group_id = None

        self.instance_url = '{}/api/v4'.format(
            self._url_match.group('base_url'))

        # fetch project_id via api, thanks to lewicki-pk
        # https://github.com/oasiswork/redmine-gitlab-migrator/pull/2
        # but also take int account, that there might be the same project in different namespaces
        path_with_namespace = (
            '{namespace}/{project_name}'.format(
                **self._url_match.groupdict()))
        projectId = -1
        groupId = None

        projects_info = self.api.get('{}/projects?owned=true'.format(self.instance_url))

        for project_attributes in projects_info:
            if project_attributes.get('path_with_namespace') == path_with_namespace:
                projectId = project_attributes.get('id')
                if project_attributes.get('namespace').get('kind') == 'group':
                    groupId = project_attributes.get('namespace').get('id')

        self.project_id = projectId
        if projectId == -1 :
            raise ValueError('Could not get project_id for path_with_namespace: {}'.format(path_with_namespace))
        if groupId:
            self.group_id = groupId

        self.api_url = (
            '{base_url}api/v4/projects/'.format(
                **self._url_match.groupdict())) + str(projectId)


    def is_repository_empty(self):
        """ Heuristic to check if repository is empty
        """
        return self.api.get(self.api_url)['default_branch'] is None

    def uploads_to_string(self, uploads):

        uploads_url = '{}/uploads'.format(self.api_url)
        l = []
        for u in uploads:

           log.info('\tuploading {} ({} / {})'.format(u['filename'], u['content_url'], u['content_type']))

           # http://docs.python-requests.org/en/latest/user/quickstart/#post-a-multipart-encoded-file
           # http://stackoverflow.com/questions/20830551/how-to-streaming-upload-with-python-requests-module-include-file-and-data

           files = []
           try:
               files = [("file", (u['filename'], urlopen(u['content_url']), u['content_type']))]
           except urllib.error.HTTPError as e:
               log.warn("{} can't upload due to error: {}!".format(u['content_url'], e))


           try:
               upload = self.api.post(
                   uploads_url, files=files)
               l.append('{} {}'.format(upload['markdown'], u['description']))

           except requests.exceptions.HTTPError:
               # gitlab might throw an "ArgumentError (invalid byte sequence in UTF-8)" in production.log
               # if the filename contains special chars like german "umlaute"
               # in that case we retry with an ascii only filename.
               try:
                   files = [("file", (self.remove_non_ascii(u['filename']), urlopen(u['content_url']), u['content_type']))]
                   upload = self.api.post(uploads_url, files=files)
                   l.append('{} {}'.format(upload['markdown'], u['description']))
               except urllib.error.HTTPError as e:
                   log.warn("{} can't upload due to error: {}!".format(u['content_url'], e))


        return "\n  * ".join(l)

    def remove_non_ascii(self, text):
        # http://stackoverflow.com/a/20078869/98491
        return ''.join([i if ord(i) < 128 else ' ' for i in text])

    def create_issue(self, data, meta, auth_header):
        """ High-level issue creation

        :param meta: dict with "sudo_user", "must_close", "notes" and "attachments" keys
        :param data: dict formatted as the gitlab API expects it
        :param auth_header: dict to with headers to auth request
        :return: the created issue (without notes)
        """

        # attachments have to be uploaded prior to creating an issue
        # attachments are not related to an issue but can be referenced instead
        # see: https://docs.gitlab.com/ce/api/projects.html#upload-a-file
        uploads_text = self.uploads_to_string(meta['uploads'])
        if len(uploads_text) > 0:
           data['description'] = "{}\n* Uploads:\n  * {}".format(data['description'], uploads_text)
        headers = auth_header
        if 'sudo_user' in meta:
            headers['SUDO'] = meta['sudo_user']
        issues_url = '{}/issues'.format(self.api_url)
        issue = None
        try:
            issue = self.api.post(
                issues_url, data=data, headers=headers)
        except requests.exceptions.HTTPError as e:
            log.error("Can't convert issue due to error: {}".format(e.response.content))
            exit()


        issue_url = '{}/{}'.format(issues_url, issue['iid'])

        # Handle issues notes
        issue_notes_url = '{}/notes'.format(issue_url, 'notes')
        for note_data, note_meta in meta['notes']:
            note_headers = auth_header
            if 'sudo_user' in note_meta:
                note_headers['SUDO'] = note_meta['sudo_user']
            self.api.post(
                issue_notes_url, data=note_data,
                headers=note_headers)

        # Handle estimated and spent time
        if meta['human_time_estimate'] is not None and meta['human_time_estimate'] != 0.0:
            time_estimate_url = '{}/time_estimate?duration={}h'.format(issue_url, meta['human_time_estimate'])
            self.api.post(time_estimate_url)
        if meta['human_total_time_spent'] is not None and meta['human_total_time_spent'] != 0.0:
            time_spent_url = '{}/add_spent_time?duration={}h'.format(issue_url, meta['human_total_time_spent'])
            self.api.post(time_spent_url)

        # Handle closed status
        if meta['must_close']:
            self.api.put(issue_url, {'state_event': 'close'})

        return issue

    def delete_issue(self, iid):
        issue_url = '{}/issues/{}'.format(self.api_url, iid)
        try:
            self.api.delete(issue_url)
        except JSONDecodeError:
            True

    def create_milestone(self, data, meta):
        """ High-level milestone creation

        :param meta: dict with "should_close"
        :param data: dict formatted as the gitlab API expects it
        :return: the created milestone
        """
        milestones_url = '{}/milestones'.format(self.api_url)

        # create milestone if not exists
        try:
            milestone = self.get_milestone_by_title(data['title'])
        except ValueError:
            milestone = self.api.post(milestones_url, data=data)

        if (meta['must_close'] and milestone['state'] != 'closed'):
            milestone_url = '{}/{}'.format(milestones_url, milestone['id'])
            altered_milestone = milestone.copy()
            altered_milestone['state_event'] = 'close'

            self.api.put(milestone_url, data=altered_milestone)
        return milestone

    def get_issues(self):
        return self.api.get('{}/issues'.format(self.api_url))

    def get_members(self):
        project_members = self.api.get('{}/members'.format(self.api_url))
        if self.group_id:
            group_members = self.get_instance().get_group_members(self.group_id)
            return project_members + group_members
        else:
            return project_members

    def get_members_index(self):
        """ Returns dict index of users (by login)
        """
        return {i['username']: i for i in self.get_members()}

    def get_milestones(self):
        if not hasattr(self, '_cache_milestones'):
            self._cache_milestones = self.api.get(
                '{}/milestones'.format(self.api_url))
        return self._cache_milestones

    def get_milestones_index(self):
        return {i['title']: i for i in self.get_milestones()}

    def get_milestone_by_id(self, _id):
        milestones = self.get_milestones()
        for i in milestones:
            if i['id'] == _id:
                return i
        raise ValueError('Could not get milestone for id {}'.format(_id))

    def get_milestone_by_title(self, _title):
        milestones = self.get_milestones()
        for i in milestones:
            if i['title'] == _title:
                return i
        raise ValueError('Could not get milestone for title {}'.format(_title))

    def has_members(self, usernames):
        gitlab_user_names = set([i['username'] for i in self.get_members()])
        return all((i in gitlab_user_names for i in usernames))

    def get_id(self):
        return self.api.get(self.api_url)['id']

    def get_instance(self):
        """ Return a GitlabInstance
        """
        return GitlabInstance(self.instance_url, self.api)




from git import Repo, Actor

import pypandoc
import logging
import re
import unicodedata

log = logging.getLogger(__name__)

class TextileConverter():
    def __init__(self):
        # make sure we use at least version 17 of pandoc
        # TODO: fix this test, it will not work properly for version 1.2 or 1.100
        version = pypandoc.get_pandoc_version()
        if (version < "1.17"):
            log.error('You need at least pandoc 1.17.0, download from http://pandoc.org/installing.html')
            exit(1)

        # precompile regular expressions
        self.regexWikiLinkWithText = re.compile(r'\\\[\\\[\s*([^\]]*?)\s*\|\s*([^\]]*?)\s*\\\]\\\]')
        self.regexWikiLinkWithoutText = re.compile(r'\\\[\\\[\s*([^\]]*?)\s*\\\]\\\]')
        self.regexTipMacro = re.compile(r'\{\{tip\((.*?)\)\}\}')
        self.regexNoteMacro = re.compile(r'\{\{note\((.*?)\)\}\}')
        self.regexWarningMacro = re.compile(r'\{\{warning\((.*?)\)\}\}')
        self.regexImportantMacro = re.compile(r'\{\{important\((.*?)\)\}\}')
        self.regexAnyMacro = re.compile(r'\{\{(.*)\}\}')
        self.regexCodeBlock = re.compile(r'\A  ((.|\n)*)', re.MULTILINE)
        self.regexCollapse = re.compile(r'({{collapse\s?\(([^)]+)\))(.*)(}})', re.MULTILINE | re.DOTALL)
        self.regexParagraph = re.compile(r'p(\(+|(\)+)?>?|=)?\.', re.MULTILINE | re.DOTALL)
        self.regexCodeHighlight = re.compile(r'(<code\s?(class=\"(.*)\")?>).*(</code>)', re.MULTILINE | re.DOTALL)
        self.regexAttachment = re.compile(r'attachment:[\'\"“”‘’„”«»](.*)[\'\"“”‘’„”«»]', re.MULTILINE | re.DOTALL)

    def wiki_link(self, match):
        name = match.group(1)
        if len(match.groups()) > 1:
            text = match.group(2)
        else:
            text = name

        name = self.normalize(name).replace(' ', '_')
        return '[{}]({})'.format(text, name)

    def normalize(self, title):
        title = title.replace("ß", "ss")
        title = title.replace("ä", "ae")
        title = title.replace("ö", "oe")
        title = title.replace("ü", "ue")
        title = unicodedata.normalize('NFD', title).encode('ascii', 'ignore').decode('ascii')
        return title

    def convert(self, text):
        text = '\n\n'.join([re.sub(self.regexCodeBlock, r'<pre>\1</pre>', block) for block in text.split('\n\n')])

        collapseResults = re.findall(self.regexCollapse, text)
        if len(collapseResults) > 0:
            for i in range(0, len(collapseResults)):
                text = text.replace(collapseResults[i][0], "<details>")
                text = text.replace(collapseResults[i][2], "<summary>{}</summary> \n\n{}".format(collapseResults[i][1], collapseResults[i][2]))
                text = text.replace(collapseResults[i][3], "</details>")
        text = re.sub(self.regexParagraph, "", text)

        # convert from textile to markdown
        try:
            text = pypandoc.convert_text(text, 'markdown_strict', format='textile')

            # pandoc does not convert everything, notably the [[link|text]] syntax
            # is not handled. So let's fix that.

            # [[ wikipage | link_text ]] -> [link_text](wikipage)
            text = re.sub(self.regexWikiLinkWithText, self.wiki_link, text, re.MULTILINE | re.DOTALL)

            # [[ link_url ]] -> [link_url](link_url)
            text = re.sub(self.regexWikiLinkWithoutText, self.wiki_link, text, re.MULTILINE | re.DOTALL)

            # nested lists, fix at least the common issues
            text = text.replace("    \\#\\*", "    -")
            text = text.replace("    \\*\\#", "    1.")

            # Redmine is using '>' for blockquote, which is not textile
            text = text.replace("&gt; ", ">")

            # wiki note macros
            text = re.sub(self.regexTipMacro, r'---\n**TIP**: \1\n---\n', text, re.MULTILINE | re.DOTALL)
            text = re.sub(self.regexNoteMacro, r'---\n**NOTE**: \1\n---\n', text, re.MULTILINE | re.DOTALL)
            text = re.sub(self.regexWarningMacro, r'---\n**WARNING**: \1\n---\n', text, re.MULTILINE | re.DOTALL)
            text = re.sub(self.regexImportantMacro, r'---\n**IMPORTANT**: \1\n---\n', text, re.MULTILINE | re.DOTALL)

            # all other macros
            text = re.sub(self.regexAnyMacro, r'\1', text, re.MULTILINE | re.DOTALL)

            # attachments in notes
            text = re.sub(self.regexAttachment, r"\n\n*(Merged from Redmine, please check first note for attachment named **\1**)*", text, re.MULTILINE | re.DOTALL)

            # code highlight
            codeHighlights = re.findall(self.regexCodeHighlight, text)
            if len(codeHighlights) > 0:
                for i in range(0, len(codeHighlights)):
                    text = text.replace(codeHighlights[i][0], "\n```{}".format(codeHighlights[i][2].lower()))
                    text = text.replace(codeHighlights[i][3], "\n```")
        except RuntimeError as e:
            return False
        return text

class WikiPageConverter():
    """
    TODO:

    * set author email address in git commit
    * adjust issue numbers in links in case they do not match ("#123")
    * check links to commits ("commit:01234abc") or changesets ("r123")
    * make all wiki pages filenames lower-case and fix all links
    * tables are sometimes not converted correctly.
    * fix anything else that pandoc does not convert correctly.

    NOTE: This was tested with pandoc 1.17.0.2 - it may not work as nice
          (or badly? :-)) with other versions.

    Redmine's Textile:
    http://www.redmine.org/projects/redmine/wiki/RedmineTextFormattingTextile
    """

    def __init__(self, local_repo_path):
        self.repo_path = local_repo_path
        self.repo = Repo(local_repo_path)

        # make sure we use at least version 17 of pandoc
        # TODO: fix this test, it will not work properly for version 1.2 or 1.100
        version = pypandoc.get_pandoc_version()
        if (version < "1.17"):
            log.error('You need at least pandoc 1.17.0, download from http://pandoc.org/installing.html')
            exit(1)

        self.textile_converter = TextileConverter()

    def convert(self, redmine_page):
        title = self.textile_converter.normalize(redmine_page["title"])
        if (title == 'Wiki'):
            title = 'home'
        print("Converting {} ({} version {})".format(title, redmine_page["title"], redmine_page["version"]))

        text = redmine_page.get('text', "")

        # create a copy of the original page (for comparison, will not be committed)
        file_name = title + ".textile"
        with open(self.repo_path + "/" + file_name, mode='wt', encoding='utf-8') as fd:
            print(text, file=fd)

        # replace some contents
        text = text.replace("{{lastupdated_at}}", redmine_page["updated_on"])
        text = text.replace("{{lastupdated_by}}", redmine_page["author"]["name"])
        text = text.replace("[[PageOutline]]", "")
        text = text.replace("{{>toc}}", "")

        text = self.textile_converter.convert(text)

        # save file with author/date
        file_name = title + ".md"
        with open(self.repo_path + "/" + file_name, mode='wt', encoding='utf-8') as fd:
            print(text.replace('\n', "\n"), file=fd)

        # todo: check for attachments
        # todo: upload attachments

        if redmine_page["comments"]:
            commit_msg = redmine_page["comments"] + " (" + title + " v" + str(redmine_page["version"]) + ")";
        else:
            commit_msg = title + ", version " + str(redmine_page["version"]);

        author = Actor(redmine_page["author"]["name"], "")
        time   = redmine_page["updated_on"].replace("T", " ").replace("Z", " +0000")

        self.repo.index.add([file_name])
        self.repo.index.commit(commit_msg, author=author, committer=author, author_date=time, commit_date=time)
